/* ----  Copyright (C) 1997 Alex Solomonoff ( asolomon@bbn.com )
This file is part of Invariant, a library of C++ tools for debugging and 
program consistancy checking. Invariant is free software; you can redistribute
it and/or modify it under the terms of the GNU Library General Public License,
version 2, as published by the Free Software Foundation.  ----- */

#ifndef _NONNULL_PTR_H_
#define _NONNULL_PTR_H_

#ifndef _ASSERT_H_
#include <invariant/assert.h>
#endif

#ifndef _INVL_CXXVERSION_H_
#	include <invariant/cxxversion.h>
#endif

// need to define ptr_diff somehow

template < class T > class
restricted_ptr
{
	typedef  int  ptr_diff;

        T * ptr;

    public:
        restricted_ptr( T * t) : ptr(t) { }
        restricted_ptr & operator=(T * t) { ptr=t; return *this; }
        restricted_ptr & operator=( const restricted_ptr & nnp )
	{	ptr = nnp.ptr; return *this;
	}

//      smart pointer stuff
        T * operator->() const { return ptr; }
        operator T* ()   const { return ptr; }

//      disallowed smart array stuff
    private:
        T & operator[](int);
        int operator-(const restricted_ptr<T> & );
        int operator-(T * );
        restricted_ptr<T> operator+(int);
        restricted_ptr<T> operator-(int);

//      this only gives a link-time error. sigh
        friend restricted_ptr<T> operator+ INVL_GUIDING_DECL (int, restricted_ptr<T> rt);
        friend restricted_ptr<T> operator- INVL_GUIDING_DECL (int, restricted_ptr<T> rt);
	friend int operator- INVL_GUIDING_DECL (T *, restricted_ptr<T> );

//      p++, ++p, p--, and --p already are not allowed.

};


template < class T > class
nonnull_ptr
{
//      No no-arg constructor allowed. must initialize with some nonzero value.
//      Obviously never compare it with NULL.

	typedef  int  ptr_diff;

        T * ptr;

    public:
        void check() const { invl_assert( ptr != ((T*)0) ); }
        nonnull_ptr( T * t) : ptr(t) { check(); }
        nonnull_ptr & operator=(T * t) { ptr=t; check(); return *this; }
        nonnull_ptr & operator=( const nonnull_ptr & nnp )
	{	ptr = nnp.ptr; check(); return *this;
	}

//      smart pointer stuff
        T * operator->() const { return ptr; }
        operator T* ()   const { return ptr; }

//      disallowed smart array stuff
    private:
        T & operator[](int);
        int operator-(const nonnull_ptr<T> & );
        int operator-(T * );
        nonnull_ptr<T> operator+(int);
        nonnull_ptr<T> operator-(int);

//      this only gives a link-time error. sigh
        friend nonnull_ptr<T> operator+ INVL_GUIDING_DECL (int, nonnull_ptr<T> rt);
        friend nonnull_ptr<T> operator- INVL_GUIDING_DECL (int, nonnull_ptr<T> rt);
	friend int operator- INVL_GUIDING_DECL (T *, nonnull_ptr<T> );
//      p++, ++p, p--, and --p already are not allowed.

};


template < class T > class
nonnull_arr_ptr
{
//      No no-arg constructor allowed. must initialize with some nonzero value.
//      Obviously never compare it with NULL.

	typedef  int  ptr_diff;

        T * ptr;

    public:
        void check() const { invl_assert( ptr != ((T*)0) ); }
        nonnull_arr_ptr( T * t) : ptr(t) { check(); }
        nonnull_arr_ptr & operator=(T * t) { ptr=t; check(); return *this; }
        nonnull_arr_ptr & operator=( const nonnull_arr_ptr & nnp )
	{	ptr = nnp.ptr; check(); return *this;
	}

//      smart pointer stuff
        T * operator->() const { return ptr; }
        operator T* ()   const { return ptr; }

//	smart pointer-to-array stuff

        nonnull_arr_ptr operator+(ptr_diff offset) const
	{	return nonnull_arr_ptr(ptr+offset); }
        nonnull_arr_ptr operator-(ptr_diff offset) const
	{	return nonnull_arr_ptr(ptr-offset); }
        nonnull_arr_ptr & operator+=(ptr_diff offset)
	{	ptr += offset; check() ; return *this; }
        nonnull_arr_ptr & operator-=(ptr_diff offset)
	{	ptr -= offset; check() ; return *this; }
        ptr_diff operator-( nonnull_arr_ptr<T> nnp) const
	{	return ptr - nnp.ptr; }

	nonnull_arr_ptr & operator++() { ++ptr; check(); return *this; } // ++ptr
	nonnull_arr_ptr & operator--() { --ptr; check(); return *this; } // --ptr

	const nonnull_arr_ptr operator++(int) // ptr++
	{	T * tmp = ptr;
		ptr++;
		check();
		return nonnull_arr_ptr(tmp);
	}
	
	const nonnull_arr_ptr operator--(int) // ptr--
	{	T * tmp = ptr;
		ptr--;
		check();
		return nonnull_arr_ptr(tmp);
	}
};


#endif //!_NONNULL_PTR_H_
