/* ----  Copyright (C) 1997 Alex Solomonoff ( asolomon@bbn.com )
This file is part of Invariant, a library of C++ tools for debugging and 
program consistancy checking. Invariant is free software; you can redistribute
it and/or modify it under the terms of the GNU Library General Public License,
version 2, as published by the Free Software Foundation.  ----- */

#ifndef _RESTRICTED_TYPES_H_
#define _RESTRICTED_TYPES_H_

// Header file for several kinds of restricted number types (templated).
// They include monotone increasing numbers, fixed sign numbers (non-negative),
// numbers restricted to an interval, and objects whose value is supposed to
// be unchanged from their starting value at certain times.

#ifndef _ASSERT_H_
#include <invariant/assert.h>
#endif

template < class type > class
unchanged
{
        type    orig;
        const type &  curr;
     public:
        unchanged( const type & t) : orig(t), curr(t) {}
        void check() const { invl_assert( curr == orig); }
        ~unchanged() { check(); }
};

// need to manually make sure that the variable to be unchanged does note
// go out of scope before the unchanged<type> object is checked or destroyed.
// Else we would get false unchange failures.
// This is not good.
// If the scope of the unchange object is contained in the scope of the object
// it guards, this is not an issue.

#ifdef __GNUC__
#	define UNCHANGED(var) unchanged<typeof(var) > __invariant__##var(var)
#	define CHECK_UNCHANGED(var) __invariant__##var.check()
#endif

// class of number that is supposed to be increasing  or decreasing  only.
template < class number, class relation> class
monotone
{
    private:
        number value;

     public:
//        monotone(const number & n) : value(n) {}
        monotone(number   n) : value(n) {}
        monotone( const monotone & mo) : value(mo.value) {}

	void check( const number & n) const
        {       relation  r;
                invl_assert( r(n, value) );
	}

	void check_prev( const number & prev_n) const
        {       relation  r;
                invl_assert( r(value, prev_n) );
	}

        monotone & operator=( const monotone & m )
        {	check(m.value);
                value = m.value;
                return *this;
        }
        monotone & operator=( const number & n )
        {	check(n);
                value = n;
                return *this;
        }
	monotone & operator=( number n ) // can't possibly need both of these.
        {	check(n);
                value = n;
                return *this;
        }

        operator number() const  { return value; }

#	ifndef __USES_PREV
#		define __USES_PREV
#	endif
#	define __ARITH_TYPE monotone
#	include <invariant/smart_num.h>
#	undef	__ARITH_TYPE
#	undef __USES_PREV
};

// class of number that is supposed to be positive or negative only.

template <class number, class relation> class
definite
{
    private:
        number value;

     public:
//        definite(number & n) : value(n) {}
        definite(number   n) : value(n) {}
        definite( const definite & d) : value(d.value) {}

	void check() const
	{	relation r;
                invl_assert( r(value, ((number)0) ) );
	}

        definite & operator=( const definite & d )
        {       value = d.value;
		check();
                return *this;
        }
        definite & operator=( const number & n )
        {       value = n;
		check();
                return *this;
        }

        definite & operator=( number n )
        {       value = n;
		check();
                return *this;
        }

	operator number() const { return value; }

#	ifdef __USES_PREV
#		undef __USES_PREV
#	endif
#	define __ARITH_TYPE definite
#	include <invariant/smart_num.h>
#	undef	__ARITH_TYPE
};


// A number class that is restricted to an interval.

template < class number, int lower, int upper > class
Interval_num
{
        number value;

    public:
        void check() const { invl_assert( value >= lower && value <= upper ); }

        Interval_num(number v) : value(v) { check(); }
        Interval_num(const Interval_num & in) : value(in.value) { check(); }

        Interval_num  & operator=( const number & p)
	{	value = p;
		check();
		return *this;
	}

        operator number() const { return value; }

#	ifdef __USES_PREV
#		undef __USES_PREV
#	endif
#	define __ARITH_TYPE Interval_num
#	include <invariant/smart_num.h>
#	undef	__ARITH_TYPE

};

#endif //!_RESTRICTED_TYPES_H_
