/* ----  Copyright (C) 1997 Alex Solomonoff ( asolomon@bbn.com )
This file is part of Invariant, a library of C++ tools for debugging and 
program consistancy checking. Invariant is free software; you can redistribute
it and/or modify it under the terms of the GNU Library General Public License,
version 2, as published by the Free Software Foundation.  ----- */

#ifndef _STL_EXTEND_H_
#define _STL_EXTEND_H_

//-------------------container predicates ---------------------

template < class Container, class Predicate >
bool
all_true( const Container & con, Predicate pred)
{
	return all_true(con.begin(), con.end(), pred);
}

template <class InputIterator, class Predicate >
bool
all_true(InputIterator first, InputIterator last, Predicate pred)
{
	for( InputIterator ii=first;   ii!=last;    ii++)
		if (!pred(*ii) ) return false;
	return true;
}


template < class Container, class Predicate >
bool
some_true( const Container & con, Predicate pred)
{
	return some_true(con.begin(), con.end(), pred);
}

template <class InputIterator, class Predicate >
bool
some_true(InputIterator first, InputIterator last, Predicate pred)
{
	for( InputIterator ii=first;   ii!=last;    ii++)
		if (pred(*ii) ) return true;
	return false;
}


template < class Container, class Predicate >
bool
exactly_one_true( const Container & con, Predicate pred)
{
	return exactly_one_true(con.begin(), con.end(), pred);
}

template <class InputIterator, class Predicate >
bool
exactly_one_true(InputIterator first, InputIterator last, Predicate pred)
{
	InputIterator ii = first;
	for( ;   ii!=last;    ii++)
		if (pred(*ii) ) goto next;
	return false;
next:
	for(++ii ;   ii!=last;    ii++)
		if (pred(*ii) ) return false;
	return true;
}

// the _c is to resolve a name ambiguity
template < class Container, class Compare >
bool
invl_is_sorted_c( const Container & con, Compare comp)
{
	return is_sorted(con.begin(), con.end(), comp);
}

template <class InputIterator, class Compare >
bool
invl_is_sorted(InputIterator first, InputIterator last, Compare comp)
{
	InputIterator prev = first;
	InputIterator curr = first; ++curr;
	for( ; curr!= last;  prev=curr, ++curr)
		if (!comp(*prev, *curr) ) return false;
	return true;
}

template <class Container >
bool
invl_is_sorted_c(const Container & con)
{
	return is_sorted(con.begin(), con.end());
}

template <class InputIterator >
bool
invl_is_sorted(InputIterator first, InputIterator last)
{
	InputIterator prev = first;
	InputIterator curr = first; ++curr;
	for( ; curr!= last;  prev=curr, ++curr)
		if ( *curr < *prev ) return false;
	return true;
}

//------------- sign predicates -----------------------

template < class T >
struct positive
{	bool operator()( const T & t ) { return t > (T)0; }
};

template < class T >
struct negative 
{	bool operator()( const T & t ) { return t < (T)0; }
};

template < class T >
struct zero 
{	bool operator()( const T & t ) { return t == (T)0; }
};

template < class T >
struct non_zero 
{	bool operator()( const T & t ) { return !(t == (T)0); }
};

template < class T >
struct non_negative 
{	bool operator()( const T & t ) { return !(t < (T)0); }
};

template < class T >
struct non_positive 
{	bool operator()( const T & t ) { return !(t > (T)0); }
};

//-------------------  sign adapters ---------------------


template < class real_pred, class T >
class positive_op
{
    private:
	real_pred rp;
    public:
	positive_op(real_pred r) : rp(r) {}
	bool operator()(const T & t ) { return rp(t) > 0; }
};

template < class real_pred, class T >
class negative_op
{
    private:
	real_pred rp;
    public:
	negative_op(real_pred r) : rp(r) {}
	bool operator()(const T & t ) { return rp(t) < 0; }
};

template < class real_pred, class T >
class zero_op
{
    private:
	real_pred rp;
    public:
	zero_op(real_pred r) : rp(r) {}
	bool operator()(const T & t ) { return rp(t) == 0; }
};

template < class real_pred, class T >
class nonnegative_op
{
    private:
	real_pred rp;
    public:
	nonnegative_op(real_pred r) : rp(r) {}
	bool operator()(const T & t ) { return rp(t) >= 0; }
};

template < class real_pred, class T >
class nonpositive_op
{
    private:
	real_pred rp;
    public:
	nonpositive_op(real_pred r) : rp(r) {}
	bool operator()(const T & t ) { return rp(t) <= 0; }
};

template < class real_pred, class T >
class nonzero_op
{
    private:
	real_pred rp;
    public:
	nonzero_op(real_pred r) : rp(r) {}
	bool operator()(const T & t ) { return !(rp(t)==0); }
};

#endif //!_STL_EXTEND_H_
