/* ----  Copyright (C) 1997 Alex Solomonoff ( asolomon@bbn.com )
This file is part of Invariant, a library of C++ tools for debugging and 
program consistancy checking. Invariant is free software; you can redistribute
it and/or modify it under the terms of the GNU Library General Public License,
version 2, as published by the Free Software Foundation.  ----- */

#ifndef _UNIT_NUMBERS_H_
#define _UNIT_NUMBERS_H_

#include <iostream.h>
#include <cmath>

#ifndef _INVL_CXXVERSION_H_
#	include <invariant/cxxversion.h>
#endif

// Header file for numbers having units ( i.e. dimensions).
// So, you can have numbers representing length (say) and
// multiply two of these to get a number representing area,
// but if you tried to assign this product to a number representing
// volume, that would not compile.

// The proxy constructor argument trick is due to
// Scott Meyers, ``More Effective C++'', Item #5.
// The purpose is to stop unwanted implicit type conversions.
template < class number> class
number_proxy
{
	number n;
    public:
	number_proxy(const number & _n) : n(_n) {}
	operator number() const { return n; }
};


template < class number, int length, int mass, int time> class
unit_number
{
    private:
        number value;
    public:
        unit_number() :value(0) {}
        unit_number(const number_proxy<number> & d) : value((number)d) {}
        unit_number(const unit_number & un) : value(un.value) {}

        unit_number operator+(const unit_number & d) const
        {       return unit_number(value+d.value);
        }
        unit_number operator-(const unit_number & d) const
        {       return unit_number(value-d.value);
        }

	unit_number operator +=(const unit_number & d)
	{	value += d.value;
		return *this;
	}

	unit_number operator -=(const unit_number & d)
	{	value -= d.value;
		return *this;
	}

	unit_number operator *=(const unit_number<number,0,0,0> & d)
	{	value *= d.convert_to_num();
		return *this;
	}

	unit_number operator /=(const unit_number<number,0,0,0> & d)
	{	value /= d.convert_to_num();
		return *this;
	}

	unit_number operator-()
	{	return unit_number(-value);
	}

        bool operator==(const unit_number & n) const
        {       return value==n.value;
        }

        bool operator>=(const unit_number & n) const
        {       return value>=n.value;
        }

        bool operator<=(const unit_number & n) const
        {       return value<=n.value;
        }

        bool operator>(const unit_number & n) const
        {       return value>n.value;
        }

        bool operator<(const unit_number & n) const
        {       return value<n.value;
        }

	number
	convert_to_num() const { return value; } // don't want this happening automatically
	operator number_proxy<number>() { return number_proxy<number>(value); }

	friend ostream & operator<<  INVL_GUIDING_DECL (ostream & os, const unit_number & dn);

};

template <class number, int l0, int l1, int m0, int m1, int t0, int t1>
unit_number<number, l0+l1, m0+m1, t0+t1>
operator*(unit_number<number, l0, m0, t0> dn0, unit_number<number, l1, m1, t1> dn1)
{	return unit_number<number, l0+l1, m0+m1, t0+t1>
				( dn0.convert_to_num()*dn1.convert_to_num() ) ;
}

template <class number, int l0, int l1, int m0, int m1, int t0, int t1>
unit_number<number, l0-l1, m0-m1, t0-t1>
operator/(unit_number<number, l0, m0, t0> dn0, unit_number<number, l1, m1, t1> dn1)
{	return unit_number<number, l0-l1, m0-m1, t0-t1>
				( dn0.convert_to_num()/dn1.convert_to_num() ) ;
}

template < class number, int length, int mass, int time> class
ostream & operator<<(ostream & os, const unit_number<number, length, mass, time> & dn)
{       os << dn.value;
        if ( length != 0)
        {       os << " * length";
		if ( length != 1) os << "^" << length;
	}
        if ( mass!= 0)
        {        os << " * mass";
		if ( mass != 1) os << "^" << mass;
	}
        if ( time!= 0)
        {        os << " * time";
		if ( time != 1) os << "^" << time;
	}
        return os;
}

#define DEF_FCN(function)					\
template < class number>					\
unit_number<number,0,0,0>					\
function( unit_number<number,0,0,0> dn)				\
{       return unit_number<number,0,0,0>( function(dn.convert_to_num()) );  \
}

DEF_FCN(sqrt)
DEF_FCN(exp)
DEF_FCN(log)
DEF_FCN(sin)
DEF_FCN(cos)
DEF_FCN(tan)
DEF_FCN(asin)
DEF_FCN(acos)
DEF_FCN(atan)
DEF_FCN(sinh)
DEF_FCN(cosh)
DEF_FCN(tanh)
DEF_FCN(asinh)
DEF_FCN(acosh)
DEF_FCN(atanh)
DEF_FCN(j0)
DEF_FCN(j1)
DEF_FCN(y0)
DEF_FCN(y1)

#undef DEF_FCN

// some macros for representing common physical quantities
// So you can write things like:
//	unit_num<double, VELOCITY>	c;
//	unit_num<double, MASS >		m; 
//	unit_num<double, ENERGY>	e = m*c*c;

#define VELOCITY		 1, 0,-1
#define RADIAL_VELOCITY		 0, 0,-1
#define PRESSURE		-2, 1,-2
#define ACCELERATION		 1, 0,-2
#define DENSITY			-3, 1, 0
#define SPECIFIC_VOLUME		 3,-1, 0
#define ENERGY			 2, 1,-2
#define POWER			 2, 1,-3
#define FORCE			 1, 1,-2
#define VOLUME			 3, 0, 0
#define VOLUME_FLUX		 3, 0,-1 
#define AREA			 2, 0, 0
#define ANGULAR_MOMENTUM	 2, 1,-1
#define MOMENTUM		 1, 1,-1
#define TORQUE			 2, 1,-2
#define LENGTH			 1, 0, 0
#define MASS			 0, 1, 0
#define TIME			 0, 0, 1
#define MASS_FLUX		 0, 1,-1
#define VISCOSITY	 	-1, 1,-1
#define KINEMATIC_VISCOSITY	 2, 0,-1
#define DIMENSIONLESS		 0, 0, 0

#endif //!_UNIT_NUMBERS_H_
